
#include "scheme.h"
#include "tacoma.h"
#include "tacoma_elk.h"


/*
 *   Variable to hold integer identifying the briefcase-type.
 */
int T_Briefcase;


/*
 *   Take a null-terminated array of strings as input.  Make
 *   a list of strings out of it, and return the list.
 */
Object make_strlist( str )
char **str;
{
    Object f = Null, p = Null, s, n;
    GC_Node4;

    /* Make_String and Cons may trigger garbage collection.  We
     * have to make f, p, s and n visible to the GC. */
    GC_Link4( f, p, s, n );

    /* Make (car,cdr)-list out of the array */
    for ( ; *str; str++ ) {
	s = Make_String( *str, strlen( *str ) );
	n = Cons( s, Null );
	if ( Nullp( p ) ) {
	    f = p = n;
	} else {
	    Cdr( p ) = n;
	    p = n;
	}
    }

    GC_Unlink;
    return f;
}


/*
 *   Check if obj is of briefcase type, and return the
 *   briefcase associated with object.
 *
 *   PS! The briefcase in obj is not copied, just returned.
 */
briefcase Get_Briefcase( obj )
Object obj;
{
    Check_Type( obj, T_Briefcase );
    return BRIEFCASE( obj )->bc;
}


/*
 *   Make a briefcase-object of bc and return it.
 *
 *   PS! The briefcase in _not_ copied into the object, and must
 *       therefore be allocated by the outside code.
 */
Object Make_Briefcase( bc )
briefcase bc;
{
    Object d = Alloc_Object( sizeof(struct s_briefcase), T_Briefcase, 0 );
    BRIEFCASE(d)->bc = bc;
    return d;
}


/*
 *   Function called to terminate briefcases when they are
 *   proven unreachable to the GC.
 */
Object Bc_Terminate( bc )
Object bc;
{
    Deregister_Object( bc );
    bc_discard( Get_Briefcase( bc ) );
    return Void;
}


/*
 *   Return true if object a and object b are identical
 *   briefcase-objects.
 */
int BC_Equal( a, b )
Object a, b;
{
    briefcase ba, bb;
    char **af, **bf, **ac, **bc;
    int i, j;

    /* First, it may be true that a and b is the same object */
    if ( BRIEFCASE(a)->bc == BRIEFCASE(b)->bc )
	return 1;

    ba = Get_Briefcase(a); af = bc_list( ba );
    bb = Get_Briefcase(b); bf = bc_list( bb );
    
    for ( i = 0; af[i] && bf[i]; i++ ) {
	/* Do not look at AGENTID-folders */
	if ( !strcmp( af[i], "AGENTID" ) && !strcmp( bf[i], "AGENTID" ) )
	    continue;

	/* Check if foldernames differ */
	if ( strcmp( af[i], bf[i] ) ) {
	    free_copy( af );
	    free_copy( bf );
	    return 0;
	}

	ac = folder_fetch( ba, af[i] );
	bc = folder_fetch( bb, bf[i] );

	for ( j = 0; ac[j] && bc[j]; j++ ) {
	    /* Check if the folders' contents differ */
	    if ( strcmp( ac[j], bc[j] ) ) {
		free_copy( af ); free_copy( bf );
		free_copy( ac ); free_copy( bc );
		return 0;
	    }
	}

	/* Check if one folder contains more elements */
	if ( ac[j] || bc[j] ) {
	    free_copy( af ); free_copy( bf );
	    free_copy( ac ); free_copy( bc );
	    return 0;
	}
	free_copy( ac );
	free_copy( bc );
    }

    /* Check if one briefcase contains more folders than the other */
    if ( af[i] || bf[i] ) {
	free_copy( af );
	free_copy( bf );
	return 0;
    }

    /* Folders are identical */
    free_copy( af );
    free_copy( bf );
    return 1;
}


/*
 *   Return true if object a and object b are the same
 *   briefcase-object.
 */
int BC_Eqv( a, b )
Object a, b;
{
    return BRIEFCASE(a)->bc == BRIEFCASE(b)->bc;
}


/*
 *   Print an object of type breifcase onto output port, port.
 */
int BC_Print( obj, port, raw, depth, length )
Object obj, port;
int raw, depth, length;
{
    briefcase b = Get_Briefcase( obj );

    Printf( port, "#[briefcase %lu]", b );
    return 0;
}


/*
**
**  Primitives for operations on briefcases.
**
*/


/*
 *   (briefcase? obj)
 *
 *   Return #t if obj is of type breifcase, #f otherwise.
 */
Object P_Briefcasep( obj )
Object obj;
{
    return TYPE(obj) == T_Briefcase ? True : False;
}


/*
 *   (bc-create)
 *
 *   Create a briefcase and return it.
 */
Object P_Bc_Create()
{
    briefcase bc = bc_create();
    Object d = Alloc_Object( sizeof(struct s_briefcase), T_Briefcase, 0 );
    BRIEFCASE(d)->bc = bc;

    /* Register object, so that bc may be freed by GC. */
    Register_Object( d, "bc", Bc_Terminate, 0 );
    return d;
}


/*
 *   (bc-discard! bc)
 *
 *   Deletes the briefcase, bc, and all folders associated with it.
 *   Returns the value #v.
 */
Object P_Bc_Discard_Set( bc )
Object bc;
{
    Deregister_Object( bc );
    bc_discard( Get_Briefcase( bc ) );
    return Void;
}


/*
 *   (bc-list bc)
 *
 *   Returns a list of all folders within briefcase, bc.
 */
Object P_Bc_List( bc )
Object bc;
{
    Object d;
    char **list;

    list = bc_list( Get_Briefcase( bc ) );
    d = make_strlist( list );
    free_copy( list );

    return d;
}


/*
 *   (archive bc)
 *   (briefcase->string bc)
 *
 *   Archives briefcase, bc, into a string and returns the string.
 */
Object P_Archive( bc )
Object bc;
{
    char *str = archive( Get_Briefcase(bc) );
    Object d = ( str == NULL ? Null : Make_String( str, strlen(str) ) );
   
    free( str );
    return d;
}


/*
 *   (unarchive! bc archive)
 *
 *   Unarchives the archive into briefcase, bc, and returns
 *   the unarchived briefcase-object or #f if error.
 */
Object P_Unarchive_Set( bc, arch )
Object bc, arch;
{
    briefcase b = unarchive( Get_Briefcase(bc), Get_Strsym(arch) );
    return b == NULL ? False : bc;
}


/*
 *   Define the briefcase-type and primitives for operations on
 *   briefcases.
 */
void elk_init_tac_briefcase()
{
    T_Briefcase = Define_Type( 0, "briefcase", 0, 
			       sizeof(struct s_briefcase),
			       BC_Eqv, BC_Equal,
			       BC_Print, 0 );
			       
    Define_Primitive( P_Briefcasep, "briefcase?", 1, 1, EVAL );
    Define_Primitive( P_Bc_Create, "bc-create", 0, 0, EVAL );
    Define_Primitive( P_Bc_Discard_Set, "bc-discard!", 1, 1, EVAL );
    Define_Primitive( P_Bc_List, "bc-list", 1, 1, EVAL );
    Define_Primitive( P_Archive, "archive", 1, 1, EVAL );
    Define_Primitive( P_Archive, "briefcase->string", 1, 1, EVAL );
    Define_Primitive( P_Unarchive_Set, "unarchive!", 2, 2, EVAL );
}
