#include "tacoma.h"
#include "tacoma_elk.h"

/*
 *   Variable to hold integer identifying the cabinet-type.
 */
int T_Cabinet;


/*
 *   Check if obj is of cabinet type, and return the
 *   cabinet associated with object.
 *
 *   PS! The cabinet in obj is not copied, just returned.
 */
cabinet Get_Cabinet( obj )
Object obj;
{
    Check_Type( obj, T_Cabinet );
    return CABINET( obj )->cab;
}


/*
 *   Function called to terminate cabinet when they are
 *   proven unreachable to the GC.
 */
Object Cab_Terminate( cab )
Object cab;
{
    Deregister_Object( cab );
    cabinet_close( Get_Cabinet( cab ) );
    free( CABINET(cab)->name );
    return Void;
}


/*
 *   Return true if object a and object b are the same
 *   cabinet-object.
 */
int Cab_Eqv( a, b )
Object a, b;
{
    return CABINET(a)->cab == CABINET(b)->cab;
}


/*
 *   Print an object of type cabinet onto output port, port.
 */
int Cab_Print( obj, port, raw, depth, length )
Object obj, port;
int raw, depth, length;
{
    cabinet cab = Get_Cabinet( obj );

    Printf( port, "#[cabinet %s%s]", CABINET(obj)->name,
	    CABINET(obj)->closed ? " (closed)" : "" );
    return 0;
}


/*
**
**  Primitives operations on cabinets.
**
*/

/*
 *   (cabinet? obj)
 *
 *   Return #t if obj is of type cabinet, #f otherwise.
 */
Object P_Cabinetp( obj )
Object obj;
{
    return TYPE(obj) == T_Cabinet ? True : False;
}


/*
 *   (cabinet-open name [create])
 *
 *   Open local cabinet, name, and return it.  If a second argument 'create
 *   is specified, create the cabinet if it does not exist.
 *   If an error occured, or cabinet did not exist (and was not created),
 *   return False.
 */
Object P_Cabinet_Open( argc, argv )
int argc;
Object *argv;
{
    cabinet cab;
    int flag = 0;
    char *name = Get_String( argv[0] );
    Object d;
    GC_Node;

    /* Should cabinet be created? */
    if ( argc == 2 ) {
	GC_Link( argv[1] );
	if ( EQ( argv[1], Intern( "create" ) ) ) 
	    flag = 1;
	GC_Unlink;
    }

    /* Open cabinet */
    if ( ( cab = cabinet_open( name, flag ) ) == NULL )
	return False;

    d = Alloc_Object( sizeof(struct s_cabinet), T_Cabinet, 0 );
    CABINET(d)->cab = cab;
    CABINET(d)->name = (char*) strdup( name );
    CABINET(d)->closed = 0;
    
    /* Register object, so that bc may be freed by GC. */
    Register_Object( d, "cab", Cab_Terminate, 0 );
    return d;
}


/*
 *   (cabinet-close! cab)
 *
 *   Closes cabinet, cab, but leaves it on local storage so that it
 *   may be opened later.
 */
Object P_Cabinet_Close_Set( cab )
Object cab;
{
    if ( !CABINET(cab)->closed ) 
	cabinet_close( Get_Cabinet( cab ) );
    CABINET(cab)->closed = 1;
    return Void;
}


/*
 *   (cabinet-discard! cab)
 *
 *   Deletes cabinet, cab, and all folders associated with it.
 */
Object P_Cabinet_Discard_Set( cab )
Object cab;
{
    if ( !CABINET(cab)->closed ) 
	cabinet_discard( Get_Cabinet(cab) );
    CABINET(cab)->closed = 1;
    return Void;
}


/*
 *   (cabinet-list cab)
 *
 *   Returns a list folders of a local cabinet, cab, or #f if error.
 */
Object P_Cabinet_List( cab )
Object cab;
{
    char **list;
    Object d;

    if ( CABINET(cab)->closed )
	return False;

    if ( (list = cabinet_list( Get_Cabinet(cab) )) == NULL )
	return False;

    d = make_strlist( list );
    free_copy( list );

    return d;
}


/*
 *   Define the cabinet-type and primitives for operations on
 *   cabinet.
 */
void elk_init_tac_cabinet()
{
    T_Cabinet = Define_Type( 0, "cabinet", 0, 
			     sizeof(struct s_cabinet),
			     Cab_Eqv, Cab_Eqv,
			     Cab_Print, 0 );

    Define_Primitive( P_Cabinetp, "cabinet?", 1, 1, EVAL );
    Define_Primitive( P_Cabinet_Open, "cabinet-open", 1, 2, VARARGS );
    Define_Primitive( P_Cabinet_Close_Set, "cabinet-close!", 1, 1, EVAL );
    Define_Primitive( P_Cabinet_Discard_Set, "cabinet-discard!", 1, 1, EVAL );
    Define_Primitive( P_Cabinet_List, "cabinet-list", 1, 1, EVAL );
}
