#include "scheme.h"
#include "tacoma.h"
#include "tacoma_elk.h"

/*
 *   Object containing the symbol named 'left'.
 */
static Object sym_left;


/*
 *   (folder-store! bc name data [data ...])
 *
 *   Store data within folder, name, located in briefcase, bc.
 *   Returns #t on success, #f if error.
 */
Object P_Folder_Store_Set( argc, argv )
int argc;
Object *argv;
{
    briefcase bc;
    char *name, **argp;
    int i, ret;
    Alloca_Begin;

    /* Get briefcase and name of folder */
    bc = Get_Briefcase( argv[0] );
    name = Get_Strsym( argv[1] );

    /* Put the rest of the arguments into an array of strings */
    Alloca( argp, char**, (argc-1) * sizeof(char *) );
    for ( i = 2; i < argc; i++ ) {
	Get_String_Stack( argv[i], argp[i-2] );
    }
    argp[i-2] = 0;

    /* Invoke folder_store within tacoma */
    ret = folder_storeA( bc, name, argp );

    Alloca_End;
    return ret == -1 ? False : True;
}


/*
 *   (folder-fetch bc name)
 *
 *   Return contents of folder, name, located in briefcase, bc, as
 *   a list of strings.
 */
Object P_Folder_Fetch( bc, f )
Object bc, f;
{
    char **list = folder_fetch( Get_Briefcase(bc), Get_Strsym(f) );
    Object d = make_strlist( list );
    
    free_copy( list );
    return d;
}


/*
 *   (folder-delete! bc name)
 *
 *   Delete contents of folder, name, located in briefcase, bc.
 *   Returns the non-printing value #v.
 */
Object P_Folder_Delete_Set( bc, f )
Object bc, f;
{
    folder_delete( Get_Briefcase(bc), Get_Strsym(f) );
    return Void;
}


/*
 *   (folder-shift! bc name dir amount [data ...])
 *
 *   Shift folder, name, located in briefcase, bc, in direction, dir
 *   (left or right) by amount, amount.  If data is supplied, shift
 *   it into the other end of the folder.
 *   Returns the last element shifted out of the folder, or the leftmost
 *   (or rightmost) element in folder if no shifting was done.
 */
Object P_Folder_Shift_Set( argc, argv )
int argc;
Object *argv;
{
    Object d;
    briefcase bc;
    char *name, **argp, *ret;
    int i, dir, amount;
    Alloca_Begin;

    /* Get briefcase, name of folder and amount to shift */
    bc = Get_Briefcase( argv[0] );
    name = Get_Strsym( argv[1] );
    amount = Get_Integer( argv[3] );

    /* Get direction of shift */
    Check_Type( argv[2], T_Symbol );
    dir = EQ( argv[2], sym_left ) ? TAC_LEFT : TAC_RIGHT;

    /* Put the rest of the arguments into an array of strings */
    Alloca( argp, char**, (argc-3) * sizeof(char *) );
    for ( i = 4; i < argc; i++ ) {
	Get_String_Stack( argv[i], argp[i-4] );
    }
    argp[i-4] = 0;

    ret = folder_shiftA( bc, name, dir, amount, argp );
    d = ret ? Make_String( ret, strlen(ret) ) : Null;
    free( ret );

    Alloca_End;
    return d;
}


/*
 *   (folder-rotate! bc name dir amount)
 *
 *   Rotate folder, name, located in briefcase, bc, in direction, dir
 *   (left or right) by amount, amount.
 *   Returns the last element rotated.
 */
Object P_Folder_Rotate_Set( bc, f, dir, amount )
Object bc, f, dir, amount;
{
    Object obj;
    char *ret;
    int d;

    /* Get direction of shift */
    Check_Type( dir, T_Symbol );
    d = EQ( dir, sym_left ) ? TAC_LEFT : TAC_RIGHT;

    ret = folder_rotate( Get_Briefcase(bc), Get_Strsym(f), d,
			 Get_Integer(amount) );
    obj = ret ? Make_String( ret, strlen(ret) ) : Null;
    free( ret );

    return obj;
}


/*
 *   Define primitives for operations on folders.
 *   Also define the symbol 'left' which is used internaly.
 */
void elk_init_tac_folder()
{
    Define_Symbol( &sym_left, "left" );

    Define_Primitive( P_Folder_Fetch, "folder-fetch", 2, 2, EVAL );
    Define_Primitive( P_Folder_Store_Set, "folder-store!", 3, MANY, VARARGS );
    Define_Primitive( P_Folder_Delete_Set, "folder-delete!", 2, 2, EVAL );
    Define_Primitive( P_Folder_Shift_Set, "folder-shift!", 4, MANY, VARARGS );
    Define_Primitive( P_Folder_Rotate_Set, "folder-rotate!", 4, 4, EVAL );
}

