/* --* Tacoma Module *--

   Author:  Kjetil Jacobsen <kjetilja@stud.cs.uit.no>

   Provides an API to the Tacoma agent-system.  Check out
     http://www.cs.uit.no/DOS/Tacoma/
   for more information on the paradigm and software support.

   Limitations/Bugs: 

   - The 'pop, push, top, add' methods have been omitted due to their
   simplicity.  Similar functionality can be given through use of the
   'shift' and 'rotate' methods.
   - Poor error-handling within methods

   Module interface:

   - tacoma.bc_create() --> Briefcase object
   - tacoma.cab_open(name, flags) --> Cabinet object

   Briefcase methods:

   - b.meet(agent, flags) --> Status (string or int)
   - b.archive() --> Archived briefcase (string)
   - b.unarchive(archive) --> Unarchives 'archive' onto briefcase (self)
   - b.list() --> List of folders in briefcase
   - b.discard() --> None
   - b.folder_store(folder, data) --> Status
   - b.folder_fetch(folder) --> List of folder data
   - b.folder_shift(folder, dir, amount, data) --> Folder element
   - b.folder_rotate(folder, dir, amount, data) --> Folder element
   - b.folder_delete(folder) --> None

   Cabinet methods:

   - c.list() --> List of folders in the cabinet
   - c.discard() --> None
   - c.close() --> None

   As given, the methods behave like their C-counterparts in Tacoma.
   Refer to the Tacoma manual pages for more detailed information.

*/

#include <stdlib.h>
#include "allobjects.h"
#include "/somewhere/tacoma/cabinets/C_files/tacoma.h"
#include "/somewhere/tacoma/cabinets/C_files/meet.h"

/* The briefcase object*/
typedef struct {
    PyObject_HEAD
    briefcase bc;
} PyBriefcaseObject;

/* The cabinet object */
typedef struct {
    PyObject_HEAD
    cabinet cab;
} PyCabinetObject;

staticforward PyTypeObject PyBriefcase_t;
staticforward PyTypeObject PyCabinet_t;

/* Need this to link with the incbc-object */
static briefcase local_bc;

/* meet() return values as strings */
static char *meetret[] = {
    "ok",
    "notreached",
    "refused",
    "nosite",
    "ag-dead",
    "ag-notfound",
    "internal"
};


/*
 * 
 * Methods for the briefcase objects
 *
 */
static PyObject *PyBriefcase_locbc(bc)
  PyBriefcaseObject *bc;
{
    /* This one should *never* be used except from
       within ag_python.c */
    if (!local_bc == NULL) {
	bc_discard(bc->bc);
	bc->bc = local_bc;
    }
    INCREF(None);
    return None;
}

static PyObject *PyBriefcase_meet(bc, args)
  PyBriefcaseObject *bc;
  object *args;
{
    char *agent;
    int flags, ret;

    if (!PyArg_Parse(args, "(si)", &agent, &flags))
	return NULL;
    ret = meet(agent, bc->bc, flags);

    return (ret < 0 ? Py_BuildValue("O", PyInt_FromLong((long) ret)):
	     Py_BuildValue("O", PyString_FromString(meetret[ret])));
}

static PyObject *PyBriefcase_archive(bc)
  PyBriefcaseObject *bc;
{
    return Py_BuildValue("O", PyString_FromString(archive(bc->bc)));
}

static PyObject *PyBriefcase_unarchive(bc, args)
  PyBriefcaseObject *bc;
  object *args;
{
    char *archive;

    if (!PyArg_Parse(args, "s", &archive))
	return NULL;

    bc->bc = (briefcase) unarchive(bc->bc, archive);
    return Py_BuildValue("O", bc);
}

static PyObject *PyBriefcase_list(bc)
  PyBriefcaseObject *bc;
{
    char **ret;
    PyListObject *list;
    int i,j;

    ret = (char **)bc_list(bc->bc);
    for (i=0; ret[i] != 0; i++);

    /* Build a list or a string depending on ret */
    if (i > 1) {
	/* List */
	list = (PyListObject *)PyList_New(i);
	for (j=0; j < i; j++)
	    list->ob_item[j] = PyString_FromString(ret[j]);
	return Py_BuildValue("O", list);
    } else
	/* String */
	return Py_BuildValue("O", PyString_FromString(ret[0]));
}   

static PyObject *PyBriefcase_discard(bc)
  PyBriefcaseObject *bc;
{
    bc_discard(bc->bc);
    PyMem_XDEL(bc->bc);
    PyMem_DEL(bc);
    Py_INCREF(Py_None);
    return Py_None;
}


/*
 *
 * Methods on folders
 *
 */
static PyObject *PyBriefcase_folder_store(bc, args)
  PyBriefcaseObject *bc;
  object *args;
{
    char *name, **gotlist;
    int i, siz, ret;
    PyListObject *mdata;
    PyObject *data;

    if (!PyArg_Parse(args, "(sO)", &name, &data ))
	return NULL;

    /* Check if the object is a list or a string */
    if (PyList_Check(data)) {
	/* List */
	mdata = (PyListObject *) data;
	siz = PyList_Size( (PyObject *) mdata);
	gotlist = malloc(sizeof(char *) * (siz+1));
	gotlist[siz] = 0;
	for (i=0; i < siz; i++)
	    gotlist[i] = PyString_AsString(mdata->ob_item[i]);
    } else {
	/* String */
	gotlist = malloc(sizeof(char *) * 2);
	gotlist[1] = 0;
	gotlist[0] = PyString_AsString(data);
    }

    ret = folder_storeA(bc->bc, name, gotlist);
    free(gotlist);
    return Py_BuildValue("i", ret);
}

static PyObject *PyBriefcase_folder_fetch(bc, args)
  PyBriefcaseObject *bc;
  object *args;
{
    char *name, **ret;
    int i, j;
    PyListObject *list;

    if (!PyArg_Parse(args, "s", &name))
	return NULL;

    ret = (char **) folder_fetch(bc->bc, name);
    for (i=0; ret[i] != 0; i++);

    /* Build a list or a string depending on ret */
    if (i > 1) {
	/* List */
	list = (PyListObject *)PyList_New(i);
	for (j=0; j < i; j++)
	    list->ob_item[j] = PyString_FromString(ret[j]);
	return Py_BuildValue("O", list);
    } else
	/* String */
	return Py_BuildValue("O", PyString_FromString(ret[0]));
}

static PyObject *PyBriefcase_folder_shift(bc, args)
  PyBriefcaseObject *bc;
  object *args;
{
    int dir, amount, i, siz;
    char *name, *ret, **gotlist;
    PyListObject *mdata;
    PyObject *data;

    if (!PyArg_Parse(args, "(siiO)", &name, &dir, &amount, &data))
	return NULL;

    /* Check if the object is a list or a string */
    if (PyList_Check(data)) {
	/* List */
	mdata = (PyListObject *)data;
	siz = PyList_Size( (PyObject *) mdata);
	gotlist = malloc(sizeof(char *) * (siz+1));
	gotlist[siz] = 0;
	for (i=0; i<siz; i++)
	    gotlist[i] = PyString_AsString(mdata->ob_item[i]);
    } else {
	/* String */
	gotlist = malloc(sizeof(char *) * 2);
	gotlist[1] = 0;
	gotlist[0] = PyString_AsString(data);
    }

    ret = (char *)folder_shiftA(bc->bc, name, dir, amount, gotlist);
    free(gotlist);
    return Py_BuildValue("O", PyString_FromString(ret));
}

static PyObject *PyBriefcase_folder_rotate(bc, args)
  PyBriefcaseObject *bc;
  object *args;
{
    int dir, amount;
    char *name, *ret;

    if (!PyArg_Parse(args, "(sii)", &name, &dir, &amount))
	return NULL;

    ret = (char *)folder_rotate(bc->bc, name, dir, amount);
    return Py_BuildValue("O", PyString_FromString(ret));
}

static PyObject *PyBriefcase_folder_delete(bc, args)
  PyBriefcaseObject *bc;
  object *args;
{
    char *name;

    if (!PyArg_Parse(args, "s", &name))
	return NULL;
    folder_delete(bc->bc, name);
    Py_INCREF(Py_None);
    return Py_None;
}


/* 
 * These folder-methods should really be done in plain python code,
 * just provided if someone wish to do them manually.
 */
static PyObject *PyBriefcase_folder_pop(bc, args)
  PyBriefcaseObject *bc;
  object *args;
{
    char *name;

    if (!PyArg_Parse(args, "s", &name))
	return NULL;

    return Py_None;
}

static PyObject *PyBriefcase_folder_push(bc, args)
  PyBriefcaseObject *bc;
  object *args;
{
    char *name;

    if (!PyArg_Parse(args, "s", &name))
	return NULL;

    return Py_None;
}

static PyObject *PyBriefcase_folder_top(bc, args)
  PyBriefcaseObject *bc;
  object *args;
{
    char *name;

    if (!PyArg_Parse(args, "s", &name))
	return NULL;

    return Py_None;
}

static PyObject *PyBriefcase_folder_add(bc, args)
  PyBriefcaseObject *bc;
  object *args;
{
    char *name;

    if (!PyArg_Parse(args, "s", &name))
	return NULL;

    return Py_None;
}


/*
 *
 * Methods for cabinet objects
 *
 */
static PyObject *PyCabinet_discard(cab)
  PyCabinetObject *cab;
{
    cabinet_discard(cab->cab);
    PyMem_XDEL(cab->cab);
    PyMem_DEL(cab);
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *PyCabinet_close(cab)
  PyCabinetObject *cab;
{
    cabinet_close(cab->cab);
    PyMem_XDEL(cab->cab);
    PyMem_DEL(cab);
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *PyCabinet_list(cab)
  PyCabinetObject *cab;
{
    char **ret;
    PyListObject *list;
    int i,j;

    ret = (char **)cabinet_list(cab->cab);
    for (i=0; ret[i] != 0; i++);

    /* Build a list or a string depending on ret */
    if (i > 1) {
	/* List */
	list = (PyListObject *)PyList_New(i);
	for (j=0; j < i; j++)
	    list->ob_item[j] = PyString_FromString(ret[j]);
	return Py_BuildValue("O", list);
    } else
	/* String */
	return Py_BuildValue("O", PyString_FromString(ret[0]));
}   


/*
 *
 * Dictionary and method tables for the objects
 *
 */

/* Briefcase methods */
static PyMethodDef PyBriefcase_methods[] = {
    {"folder_store", (PyCFunction)PyBriefcase_folder_store},
    {"folder_fetch", (PyCFunction)PyBriefcase_folder_fetch},
    {"folder_delete",(PyCFunction)PyBriefcase_folder_delete},
    {"folder_rotate",(PyCFunction)PyBriefcase_folder_rotate},
    {"folder_shift", (PyCFunction)PyBriefcase_folder_shift},
    {"folder_pop",    (PyCFunction)PyBriefcase_folder_pop},
    {"folder_push",  (PyCFunction)PyBriefcase_folder_push},
    {"folder_add",   (PyCFunction)PyBriefcase_folder_add},
    {"folder_top",   (PyCFunction)PyBriefcase_folder_top},
    {"archive",       (PyCFunction)PyBriefcase_archive},
    {"unarchive",     (PyCFunction)PyBriefcase_unarchive},
    {"list",          (PyCFunction)PyBriefcase_list},
    {"discard",       (PyCFunction)PyBriefcase_discard},
    {"meet",          (PyCFunction)PyBriefcase_meet},
    {"locbc",         (PyCFunction)PyBriefcase_locbc},
    {NULL, NULL}
};

/* Cabinet methods */
static PyMethodDef PyCabinet_methods[] = {
    {"close",       (PyCFunction)PyCabinet_close},
    {"discard",     (PyCFunction)PyCabinet_discard},
    {"list",        (PyCFunction)PyCabinet_list},
    {NULL, NULL}
};


static PyObject *PyBriefcase_Getattr(bc, name)
  PyBriefcaseObject *bc;
  char *name;
{
    return Py_FindMethod(PyBriefcase_methods, (PyObject *)bc, name);
}

static void PyBriefcase_Dealloc(bc)
   PyBriefcaseObject *bc;
{
    bc_discard(bc->bc);
    PyMem_XDEL(bc->bc);
    PyMem_DEL(bc);
}

static PyObject *PyCabinet_Getattr(cab, name)
  PyCabinetObject *cab;
  char *name;
{
    return Py_FindMethod(PyCabinet_methods, (PyObject *)cab, name);
}

static void PyCabinet_Dealloc(cab)
   PyCabinetObject *cab;
{
    cabinet_close(cab->cab);
    PyMem_XDEL(cab->cab);
    PyMem_DEL(cab);
}

/* Briefcase type */
static PyTypeObject PyBriefcase_t = {
        OB_HEAD_INIT(&Typetype)
        0,                        /*ob_size*/
        "PyBriefcase",           /*tp_name*/
        sizeof(PyBriefcaseObject),/*tp_basicsize*/
        0,                         /*tp_itemsize*/
        /* methods */
        (destructor)PyBriefcase_Dealloc, /*tp_dealloc*/
        0,                               /*tp_print*/
        (getattrfunc)PyBriefcase_Getattr, /*tp_getattr*/
        0,                      /*tp_setattr*/
        0,                      /*tp_compare*/
        0,                      /*tp_repr*/
        0,                      /*tp_as_number*/
        0,                      /*tp_as_sequence*/
        0,                      /*tp_as_mapping*/
        0,                      /*tp_hash*/
};

/* Cabinet type */
static PyTypeObject PyCabinet_t = {
        OB_HEAD_INIT(&Typetype)
        0,                        /*ob_size*/
        "PyCabinet",             /*tp_name*/
        sizeof(PyCabinetObject),  /*tp_basicsize*/
        0,                         /*tp_itemsize*/
        /* methods */
        (destructor)PyCabinet_Dealloc,  /*tp_dealloc*/
        0,                              /*tp_print*/
        (getattrfunc)PyCabinet_Getattr, /*tp_getattr*/
        0,                      /*tp_setattr*/
        0,                      /*tp_compare*/
        0,                      /*tp_repr*/
        0,                      /*tp_as_number*/
        0,                      /*tp_as_sequence*/
        0,                      /*tp_as_mapping*/
        0,                      /*tp_hash*/
};

/* Create a new briefcase object */
static PyBriefcaseObject *tacoma_bc_create()
{
    PyBriefcaseObject *bc;

    bc = PyObject_NEW(PyBriefcaseObject, &PyBriefcase_t);
    bc->bc = (briefcase)bc_create();

    return bc;
}

/* Create a cabinet object, or return an existing one */
static PyCabinetObject *tacoma_cab_open(self, args)
  object *self;
  object *args;

{
    PyCabinetObject *cab;
    cabinet c_cab;
    char *name;
    int flags = 0;

    if (!PyArg_Parse(args, "(si)", &name, &flags))
	return NULL;

    if ((c_cab = cabinet_open(name, flags)) == NULL) {
	Py_INCREF(Py_None);
	return (PyCabinetObject *) Py_None;
    }

    cab = PyObject_NEW(PyCabinetObject, &PyCabinet_t);
    cab->cab = c_cab;
    return cab;
}


/* Methods table for the module */
static struct methodlist TacomaMethods[] = {
    {"bc_create", (PyCFunction)tacoma_bc_create},
    {"cab_open",  (PyCFunction)tacoma_cab_open},
    {NULL, NULL}
};


/* Tie incoming bc to a named object, used in ag_python.c */
void tietolocal(briefcase bc)
{
    local_bc = bc;
}

/* Insert 'int' value into dictionary */
static void insint(d, name, value)
  object *d;
  char *name;
  int value;
{
    object *v = PyInt_FromLong((long) value);
    if (v == NULL) {
	PyErr_Clear();
    }
    else {
	PyDict_SetItemString(d, name, v);
	Py_DECREF(v);
    }
}

/*
 * Module initialization
 */
void inittacoma()
{
    object *m, *d;

    /* Initialize the module */
    m = Py_InitModule("tacoma", TacomaMethods);
    d = PyModule_GetDict(m);

    /* Meet flags */
    insint(d, "sync", TM_SYNC);
    insint(d, "rpc", TM_SYNC);
    insint(d, "local", TM_LOCL);
    insint(d, "nomerge", TM_NMRG);
    insint(d, "ignerr", TM_IERR);
    insint(d, "crypt", TM_CRPT);
    insint(d, "sign", TM_SIGN);
    insint(d, "path", TM_PATH);
    insint(d, "nolog", TM_NLOG);

    /* Direction flags (folder_shift etc.) */
    insint(d, "left", TAC_LEFT);
    insint(d, "right", TAC_RIGHT);

    /* cabinet_open() flags */
    insint(d, "none", TC_NONE);
    insint(d, "create", TC_CREAT);
    
    /* Nullify the local briefcase pointer */
    local_bc = NULL;
}
