#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "tacoma.h"


static void	set_program( int, char*[] );
static void	print_desc( int );
static int	parse_cmdline( int, string [] );
static void	cmd_list( cabinet, int, int, string [] );
static void	cmd_add( cabinet, int, int, string [] );
static void	cmd_del( cabinet, int, int, string [] );
static void	cmd_flush( cabinet, int, int, string [] );
static void	cmd_change( cabinet, int, int, string [] );


static enum { HOST, PRINCIPAL, AGENT }	prog;
static string	sysfolder;
static string	program;

static void* cmd_func[]  = {
    (void*)cmd_list,
    (void*)cmd_add,
    (void*)cmd_del,
    (void*)cmd_flush,
    (void*)cmd_change,
    NULL
};

static string   commands[] = {
    "list",
    "add",
    "delete",
    "flush",
    "change",
    NULL };

static string	message[][3] = {
    { "host", "principal", "agent" },
    { "autoaccepted", "accepted", "autostarted" },
    { "Autoaccepted", "Accepted", "Autostarted" }
};


int
main( int argc, char* argv[] )
{
    int	     choice;
    string   tacomapath;
    cabinet  syscab;

    /* set variables which depend on program name */
    set_program(argc, argv);

    /* get and check environment variable */
    if ( (tacomapath = getenv(TAC_ENV_PATH)) == NULL ) {
	fprintf(stderr, "%s: Environment variable '" TAC_ENV_PATH \
		"' not set.\n", program);
	exit(1);
    }

    /* open system cabinet */
    if (  (syscab = cabinet_open(TAC_SYSTEM_CAB, TC_CREAT)) == NULL ) {
	fprintf(stderr, "\n%s: Could not access system cabinet: %s\n",
		program, strerror(errno));
	exit(1);
    }

    /* parse command line */
    choice = parse_cmdline( argc, argv );

    /* do commands */
    if ( choice >= 0 )
	((void (*)()) cmd_func[choice])(syscab, choice, argc, argv);

    cabinet_close(syscab);

    return (0);
}


static void
set_program( int argc, char* argv[] )
{
    /* remember program name */
    for ( program=argv[0]; strchr(program, '/'); program++ )
	program = strchr(program, '/');
    program = strdup(program);

    /* find out _which_ folder to work on */
    if ( strcmp(program, "tac_autohosts") == 0 ) {
	prog = HOST;
	sysfolder = TAC_AUTOHOSTS;
    }
    else if ( strcmp(program, "tac_acceptfrom") == 0 ) {
	prog = PRINCIPAL;
	sysfolder = TAC_ACCEPTFROM;
    }
    else if ( strcmp(program, "tac_startup") == 0 ) {
	prog = AGENT;
	sysfolder = TAC_STARTUP_AG;
    }
    else {
	fprintf(stderr, "\n%s: Not a valid start name\n\n", program);
	exit(1);
    }
}

static void
print_desc( int choice )
{
    switch ( choice ) {
      case 0:
	fprintf(stderr, "List all currently %s %ss.\n",
		message[1][prog], message[0][prog]);
	break;
      case 1:
	fprintf(stderr, "Add %s(s) to list of %s %ss.\n",
		message[0][prog], message[1][prog], message[0][prog]);
	break;
      case 2:
	fprintf(stderr, "Deletes %s(s) from list of %s %ss.\n",
		message[0][prog], message[1][prog], message[0][prog]);
	break;
      case 3:
	fprintf(stderr, "Deletes all %s %ss.\n",
		message[1][prog], message[0][prog]);
	break;
      case 4:
	fprintf(stderr, "Change one %s %s to another.\n",
		message[1][prog], message[0][prog]);
	break;
    }
}

static int
parse_cmdline( int argc, string argv[] )
{
    int	     idx;
    int	     choice = -1;

    if ( argc > 1 )
	for ( idx=0; commands[idx] != NULL; idx++ )
	    if ( strcmp(argv[1], commands[idx]) == 0 ) {
		choice = idx;
		break;
	    }

    if ( choice < 0 ) {
	fprintf(stderr, "\nUsage: %s cmd [arg(s)s]\n", program);
	fprintf(stderr, "    where cmd is one of:\n");
	for ( idx=0; commands[idx]; idx++ ) {
	    fprintf(stderr, "\t%s\t", commands[idx]);
	    print_desc(idx);
	}
	fprintf(stderr, "\n");
	exit(1);
    }

    return (choice);
}

static void
cmd_list( cabinet syscab, int choice, int argc, string argv[] )
{
    string*  str_p;
    string*  strings;

    if ( (strings = folder_fetch(syscab, sysfolder)) == NULL ) {
	printf("\nNo %ss in system folder '%s'.\n\n", message[0][prog], sysfolder);
	return;
    }

    printf("\n%s %ss:\n", message[2][prog], message[0][prog]);
    for ( str_p=strings; *str_p; str_p++ )
	printf("    %s\n", *str_p);
    printf("\n");
}

static void
cmd_add( cabinet syscab, int choice, int argc, string argv[] )
{
    int	     i;
    int	     idx;
    int	     count;
    int	     found;
    string*  str_p;
    string*  strings;

    if ( argc < 3 ) {
	fprintf(stderr, "\nUsage: %s %s <%s> [<%s(s)>]\n    ", program,
		commands[choice], message[0][prog], message[0][prog]);
	print_desc(choice);
	fprintf(stderr, "\n");
	return;
    }

    if ( (strings = folder_fetch(syscab, sysfolder)) == NULL )
	count = 0;
    else
	for ( count=0, str_p=strings; *str_p; count++, str_p++ )
	    ;

    for ( idx=2; idx < argc; idx++ ) {
	for ( found=0, i=0; i<count; i++ )
	    if ( strcmp(strings[i], argv[idx]) == 0 ) {
		printf("%s already in list of %s %ss.\n", argv[idx],
		       message[1][prog], message[0][prog]);
		found = 1;
		break;
	    }
	if ( found )
	    continue;

	if (strings) {
	    strings = realloc(strings, (count+2)*sizeof(string));
	} else {
	    strings = malloc( (count+2)*sizeof(string));
	}
	if ( !strings ) {
	    fprintf(stderr, "\nCould not allocate memory!\n\n");
	    return;
	}
	strings[count++] = strdup(argv[idx]);
    }
    strings[count] = NULL;

    folder_storeA(syscab, sysfolder, strings);
    free_copy(strings);

    return;
}

static void
cmd_del( cabinet syscab, int choice, int argc, string argv[] )
{
    int	     idx;
    int	     count;
    string*  str_p;
    string*  strings;
    string*  data;

    if ( argc < 3 ) {
	fprintf(stderr, "\nUsage: %s %s <%s> [<%s(s)>]\n    ", program,
		commands[choice], message[0][prog], message[0][prog]);
	print_desc(choice);
	fprintf(stderr, "\n");
    }

    if ( (strings = folder_fetch(syscab, sysfolder)) == NULL ) {
	printf("\nNo %ss in system folder '%s'.\n\n", message[0][prog], sysfolder);
	return;
    }

    for ( count=0, str_p=strings; *str_p; count++, str_p++ )
	;

    for ( idx=2; idx < argc; idx++ )
	for ( str_p=strings; *str_p; str_p++ )
	    if ( strcmp(argv[idx], *str_p) == 0 ) {
		free(*str_p);
		*str_p = strdup("");
		count--;
	    }

    if ( (data = malloc((count+1)*sizeof(string))) == NULL ) {
	fprintf(stderr, "\nCould not allocate memory!\n\n");
	free_copy(strings);
	return;
    }

    for ( idx=0, str_p=strings; *str_p; str_p++ )
	if ( **str_p )
	    data[idx++] = strdup(*str_p);
    data[idx] = NULL;

    folder_storeA(syscab, sysfolder, data);
    free_copy(strings);
    free_copy(data);

    return;
}

static void
cmd_flush( cabinet syscab, int choice, int argc, string argv[] )
{
    folder_store(syscab, sysfolder, "", NULL);
}

static void
cmd_change( cabinet syscab, int choice, int argc, string argv[] )
{
    string*  str_p;
    string*  strings;

    if ( argc != 4 ) {
	fprintf(stderr, "\nUsage: %s %s <old%s> <new%s>\n    ", program,
		commands[choice], message[0][prog], message[0][prog]);
	print_desc(choice);
	fprintf(stderr, "\n");
	return;
    }

    if ( (strings = folder_fetch(syscab, sysfolder)) == NULL ) {
	printf("\nNo %ss in system folder '%s'.\n\n", message[0][prog], sysfolder);
	return;
    }

    for ( str_p=strings; *str_p; str_p++ )
	if ( strcmp(argv[3], *str_p) == 0 ) {
	    fprintf(stderr, "\n%s already in list of %s %ss.\n\n", argv[3],
		    message[1][prog], message[0][prog]);
	    free_copy(strings);
	    exit(1);
	}

    for ( str_p=strings; *str_p; str_p++ )
	if ( strcmp(argv[2], *str_p) == 0 ) {
		free(*str_p);
		*str_p = strdup(argv[3]);
	}

    folder_storeA(syscab, sysfolder, strings);
    free_copy(strings);

    return;
}
